import time
import redis
from cobweb import Seed, log
from cobweb.decorators import check_redis_status


class RedisDB:

    def __init__(
            self,
            project: str,
            task_name: str,
            config: dict
    ):
        pool = redis.ConnectionPool(**config)
        self.heartbeat_key = f"{project}:{task_name}:heartbeat"  # redis type string
        self.spider_key = f"{project}:{task_name}:seed_info:spider"  # redis type zset, .format(priority)
        self.storer_key = f"{project}:{task_name}:seed_info:storer:%s"  # redis type set,
        self.failed_key = f"{project}:{task_name}:seed_info:failed"  # redis type set, .format(priority)
        self.succeed_key = f"{project}:{task_name}:seed_info:succeed"  # redis type set, .format(priority)
        self.update_lock = f"{project}:{task_name}:update_seed_lock"  # redis type string
        self.check_lock = f"{project}:{task_name}:check_seed_lock"  # redis type string
        self.scheduler_lock = f"{project}:{task_name}:scheduler_lock"  # redis type string
        self.client = redis.Redis(connection_pool=pool)
    
    @check_redis_status
    def _get_lock(self, key, t=15, timeout=3, sleep_time=0.1):
        begin_time = int(time.time())
        while True:
            if self.client.setnx(key, ""):
                self.client.expire(key, t)
                return True
            if int(time.time()) - begin_time > timeout:
                break
            time.sleep(sleep_time)

        if self.client.ttl(key) == -1:
            delete_status = True
            for _ in range(3):
                if self.client.ttl(key) != -1:
                    delete_status = False
                    break
                time.sleep(0.5)
            if delete_status:
                self.client.expire(key, t)
            return False
        else:
            ttl = self.client.ttl(key)
            log.info("ttl: " + str(ttl))
            return False

    @check_redis_status
    def _deal_seed(self, seeds, is_add: bool):
        if not seeds:
            return None

        if not isinstance(seeds, list):
            seeds = [seeds]

        item_info = dict()

        for seed in seeds:
            if not isinstance(seed, Seed):
                seed = Seed(seed)
            item_info[seed.format_seed] = seed._priority

        if item_info:
            self.client.zadd(self.spider_key, mapping=item_info, nx=is_add, xx=not is_add)

    @check_redis_status
    def add_seed(self, seeds):
        self._deal_seed(seeds, is_add=True)

    @check_redis_status
    def reset_seed(self, seeds):
        self._deal_seed(seeds, is_add=False)

    @check_redis_status
    def del_seed(self, seeds, spider_status: bool = True):
        if not seeds:
            return None

        if not isinstance(seeds, list):
            seeds = [seeds]

        seeds = [seed if isinstance(seed, Seed) else Seed(seed) for seed in seeds]

        if seeds:
            redis_key = self.succeed_key if spider_status else self.failed_key
            self.client.sadd(redis_key, *(str(seed) for seed in seeds))
            self.client.zrem(self.spider_key, *(seed.format_seed for seed in seeds))

    @check_redis_status
    def set_storer(self, key, seeds):
        if not seeds:
            return None

        if not isinstance(seeds, list):
            seeds = [seeds]

        item_info = dict()
        score = -int(time.time())
        for seed in seeds:
            if not isinstance(seed, Seed):
                seed = Seed(seed)
            item_info[seed.format_seed] = score

        if item_info:
            self.client.zadd(self.storer_key % key, mapping=item_info)
            log.info(f"zadd storer key: length {len(item_info.keys())}")

    @check_redis_status
    def get_seed(self, length: int = 200):
        cs = time.time()

        if self._get_lock(key=self.update_lock):

            update_item, result = {}, []

            version = int(time.time())

            items = self.client.zrangebyscore(self.spider_key, min=0, max="+inf", start=0, num=length, withscores=True)

            for value, priority in items:
                score = -(version + int(priority) / 1000)
                seed = Seed(value, priority=priority, version=version)
                update_item[value] = score
                result.append(seed)

            log.info("set seeds into queue time: " + str(time.time() - cs))
            if result:
                self.client.zadd(self.spider_key, mapping=update_item, xx=True)

            self.client.delete(self.update_lock)
            log.info("push seeds into queue time: " + str(time.time() - cs))
            return result

    @check_redis_status
    def check_spider_queue(self, stop, storer_num):
        while not stop.is_set():
            # 每15s获取check锁,等待600s后仍获取不到锁则重试;获取到锁后，设置锁的存活时间为15s
            if self._get_lock(key=self.check_lock, t=15, timeout=600, sleep_time=3):
                heartbeat = True if self.client.exists(self.heartbeat_key) else False
                # 重启重制score值，否则获取10分钟前的分数值
                score = -int(time.time()) + 600 if heartbeat else "-inf"

                keys = self.client.keys(self.storer_key % "*")
                if len(keys) == storer_num:
                    intersection_key = self.storer_key % "intersection"
                    self.client.delete(intersection_key)
                    self.client.zinterstore(intersection_key, keys)
                    while True:
                        members = self.client.zrange(intersection_key, 0, 1999)
                        if not members:
                            break
                        for key in keys:
                            self.client.zrem(key, *members)
                        self.client.sadd(self.succeed_key, *members)
                        self.client.zrem(self.spider_key, *members)
                        self.client.zrem(intersection_key, *members)
                        log.info("succeed spider data ...")

                for key in keys:
                    self.client.zremrangebyscore(key, min=score, max="(0")

                while True:
                    items = self.client.zrangebyscore(self.spider_key, min=score, max="(0", start=0, num=5000, withscores=True)
                    if not items:
                        break
                    reset_items = {}
                    for value, priority in items:
                        reset_score = "{:.3f}".format(priority).split(".")[1]
                        reset_items[value] = int(reset_score)
                    if reset_items:
                        self.client.zadd(self.spider_key, mapping=reset_items, xx=True)

                if not heartbeat:
                    self.client.setex(self.heartbeat_key, 15, "")

                self.client.delete(self.check_lock)
                time.sleep(3)

    @check_redis_status
    def set_heartbeat(self, stop):
        time.sleep(5)
        while not stop.is_set():
            self.client.setex(self.heartbeat_key, 5, "")
            time.sleep(3)

    # @check_redis_status
    # def heartbeat(self):
    #     """
    #     返回心跳key剩余存活时间
    #     """
    #     return self.client.ttl(self.heartbeat_key)

    @check_redis_status
    def spider_queue_length(self):
        return self.client.zcard(self.spider_key)

    @check_redis_status
    def ready_seed_length(self):
        return self.client.zcount(self.spider_key, min=0, max="+inf")

    @check_redis_status
    def get_scheduler_lock(self):
        return self._get_lock(self.scheduler_lock)



import oss2
from typing import Union
from oss2.models import PartInfo
from requests import Response
from cobweb import log


class OssDB:

    def __init__(
            self,
            bucket_name,
            endpoint,
            access_key,
            secret_key,
            chunk_size,
            min_size
    ):
        self.endpoint = endpoint
        self.bucket_name = bucket_name
        self.auth = oss2.Auth(
            access_key_id=access_key,
            access_key_secret=secret_key
        )
        self.bucket = oss2.Bucket(
            auth=self.auth,
            endpoint=endpoint,
            bucket_name=bucket_name
        )
        self.chunk_size = chunk_size or 1024 ** 2
        self.min_size = min_size or 1024

    @staticmethod
    def format_upload_len(length):
        if not length:
            raise ValueError("Length cannot be None or 0")

        units = ["KB", "MB", "GB", "TB"]
        for i in range(3):
            num = length / 1024 ** (i + 1)
            if num >= 1:
                return f"{round(num, 2)} {units[i]}"

    def assemble(self, ready_data, part_data):
        upload_data = None
        ready_data = ready_data + part_data
        if len(ready_data) >= self.chunk_size:
            upload_data = ready_data[:self.chunk_size]
            ready_data = ready_data[self.chunk_size:]

        return ready_data, upload_data

    def iter_data(self, data):
        if isinstance(data, Response):
            for part_data in data.iter_content(self.chunk_size):
                yield part_data
        if isinstance(data, bytes):
            for i in range(0, len(data), self.chunk_size):
                yield data[i:i + self.chunk_size]

    def upload_split(
            self, oss_path: str,
            data: Union[bytes, Response],
            timeout: int = 300,
    ):
        parts = []
        status = False
        upload_id = None
        ready_data = b""
        upload_data_len = 0
        headers = {"Expires": str(timeout * 1000)}
        try:
            upload_id = self.bucket.init_multipart_upload(oss_path).upload_id
            for part_data in self.iter_data(data):
                upload_data_len += len(part_data)
                ready_data, upload_data = self.assemble(ready_data, part_data)
                if upload_data:
                    part_index = len(parts) + 1
                    upload_info = self.bucket.upload_part(
                        oss_path, upload_id, part_index, upload_data
                    )
                    parts.append(PartInfo(part_index, upload_info.etag))

            format_upload = self.format_upload_len(upload_data_len)

            if parts and ready_data:
                part_index = len(parts) + 1
                upload_info = self.bucket.upload_part(
                    oss_path, upload_id, part_index, ready_data
                )
                parts.append(PartInfo(part_index, upload_info.etag))
                self.bucket.complete_multipart_upload(
                    oss_path, upload_id, parts
                )
                log.info(
                    f"split upload, file path: {oss_path}"
                    f", file size: {format_upload}"
                )

            elif len(ready_data) > self.min_size:
                self.bucket.put_object(oss_path, ready_data, headers)
                log.info(
                    f"upload file, file path: {oss_path}"
                    f", file size: {format_upload}"
                )

            else:
                log.info(
                    f"file size smaller than min size! "
                    f"file size: {format_upload}"
                )
            status = True
        except oss2.exceptions.RequestError as e:
            self.bucket = oss2.Bucket(
                auth=self.auth,
                endpoint=self.endpoint,
                bucket_name=self.bucket_name
            )
            log.exception("oss timeout! " + str(e))
        except Exception as e:
            self.bucket.abort_multipart_upload(oss_path, upload_id, headers)
            log.exception("upload file exception: " + str(e))

        return status




from .bbb import Seed, Queue, DBItem
from .task import Task
from .log import log
from .interface import SchedulerInterface, StorerInterface
from .db.redis_db import RedisDB
from .db.oss_db import OssDB
from .distributed.launcher import launcher





import time
from hashlib import md5
from cobweb import log, Queue, Seed, StorerInterface, SchedulerInterface
# from pympler import asizeof


class Scheduler:

    def schedule_seed(self, ready_seed_length, get_scheduler_lock, add_seed):

        if not issubclass(self.__class__, SchedulerInterface):
            raise Exception("not have schedule function!")

        if self.__class__.__name__ == "Default":
            self.stop = True
            return None

        while not self.stop:
            length = ready_seed_length()
            if length > self.size:
                time.sleep(15)

            elif get_scheduler_lock():
                seeds = self.schedule()
                add_seed(seeds)

        log.info(f"close thread: schedule_seed")

    def schedule_task(self, stop, get_seed, ready_seed_length):
        time.sleep(3)
        while not stop.is_set():

            if not ready_seed_length():
                time.sleep(15)
                continue

            if self.queue.length >= self.length:
                time.sleep(3)
                continue

            seeds = get_seed(self.length)
            self.queue.push(seeds)
        log.info(f"close thread: schedule_task")


class Spider:

    def __init__(self, queue, max_retries=5):
        self.spider_in_progress = Queue()
        self.max_retries = max_retries
        self.queue = queue

    def spider_task(self, stop, func, item, del_seed):
        while not stop.is_set():
            seed = self.queue.pop()
            if not seed:
                time.sleep(3)
                continue
            elif seed._retry > self.max_retries:
                del_seed(seed, spider_status=False)
                continue
            try:
                self.spider_in_progress.push(1)
                # log.info("spider seed: " + str(seed))
                status = None
                store_queue = None
                store_data = list()
                for it in func(item, seed):
                    if getattr(it, "table_name", None):
                        if not store_queue:
                            store_queue = it.queue()
                        store_data.append(it.struct_data)
                    elif isinstance(it, Seed):
                        self.queue.push(it)
                    elif any(isinstance(it, t) for t in (list, tuple)):
                        self.queue.push([s if isinstance(s, Seed) else Seed(s) for s in it])
                    elif isinstance(it, bool):
                        status = it

                if store_queue and store_data:
                    store_data.append(seed)
                    store_queue.push(store_data)

                if status is not None:
                    if status:
                        del_seed(seed, spider_status=True)
                    else:
                        seed._retry += 1
                        self.queue.push(seed)

            except Exception as e:
                seed._retry += 1
                self.queue.push(seed)
                log.info(f"{str(seed)} -> {str(e)}")
            finally:
                self.spider_in_progress.pop()
        log.info(f"close thread: spider")


class Storer:

    def store_task(self, stop, last, reset_seed, set_storer):

        if not issubclass(self.__class__, StorerInterface):
            return None

        if not getattr(self, "store", None):
            raise Exception("not have store function!")

        storer_name = self.__class__.__name__ + self.table
        store_key_id = md5(storer_name.encode()).hexdigest()

        while not stop.is_set():

            if last.is_set() or self.queue.length > self.length:
                seeds, data_list = [], []

                while True:
                    data = self.queue.pop()
                    if not data:
                        break
                    if isinstance(data, Seed):
                        seeds.append(data)
                        if len(data_list) >= self.length:
                            break
                        continue
                    data_list.append(data)

                if data_list:
                    if self.store(data_list):
                        set_storer(store_key_id, seeds)
                    else:
                        reset_seed(seeds)
                    continue

            time.sleep(3)

        log.info(f"close thread: {storer_name}")



from cobweb import log, Seed, SchedulerInterface


class Textfile(SchedulerInterface):

    index = None

    def schedule(self):
        try:
            seeds = []
            with open(self.table, "r") as fp:
                fp.seek(self.index or 0, 0)
                for _ in range(self.length):
                    data = fp.readline().strip()
                    if not data:
                        log.info("scheduler end!")
                        self.stop = True
                        break
                    seeds.append(Seed(data))
                    self.index = fp.tell()
            return seeds
        except FileNotFoundError:
            log.error("task table not found!")
            return None
        except TypeError:
            log.error("task table type error!")
            return None



import time
# from pympler import asizeof
from single.nest import Queue
from single.nest import struct_queue_name
from single.nest import SchedulerInterface, StorerInterface


# class Transceiver:
class Distributor:

    def __init__(self):
        self.seed_queue = Queue()

    @property
    def queue_names(self):
        return tuple(self.__dict__.keys())

    @property
    def used_memory(self):
        return asizeof.asizeof(self)

    def create_queue(self, queue_name: str):
        self.__setattr__(queue_name, Queue())

    def get_queue(self, queue_name: str):
        return self.__getattribute__(queue_name)

    def deal_item(self, item):
        icn = item.__class__.__name__
        if icn == "Seed":
            self.seed_queue.push(item)
        elif getattr(item, "table_name", None):
            queue_name = struct_queue_name(icn, item.table_name)
            getattr(self, queue_name).push(item.serialization)

    def distribute(self, callback, *args, **kwargs):
        iterable = callback(*args, **kwargs)
        if not iterable:
            return None
        for result in iterable:
            self.deal_item(result)
        return True


class Scheduler:

    def schedule_task(self, distribute):

        if not issubclass(self.__class__, SchedulerInterface):
            return None

        if not getattr(self, "schedule", None):
            raise Exception("not have schedule function!")

        while not self.stop:

            if self.queue.length < self.length:
                distribute(self.schedule)

            else:
                print("------------")
                time.sleep(15)


class Spider:

    def __init__(self, queue):
        self.queue = queue
        self.spider_in_progress = Queue()

    def spider_task(self, stop_event, distribute, func, item):
        while not stop_event.is_set():
            seed = self.queue.pop()
            if not seed:
                time.sleep(3)
                continue
            try:
                self.spider_in_progress.push(1)
                distribute(func, item, seed)
            except Exception as e:
                print(e)
            finally:
                self.spider_in_progress.pop()


class Storer:

    def store_task(self, stop_event, last_event, distribute):

        if not issubclass(self.__class__, StorerInterface):
            return None

        if not getattr(self, "store", None):
            raise Exception("not have store function!")

        while not stop_event.is_set():
            if last_event.is_set() or self.queue.length > self.length:
                data_list = []
                data_length = min(self.queue.length, self.length)
                for _ in range(data_length):
                    data = self.queue.pop()
                    data_list.append(data)
                if data_list:
                    distribute(self.store, data_list)



from .utils import parse_info, struct_start_seeds


class Task:

    def __init__(
            self,
            seeds=None,
            project=None,
            task_name=None,
            oss_config=None,
            redis_info=None,
            storer_info=None,
            scheduler_info=None,
            spider_num=None,
            max_retries=None,
            storer_queue_length=None,
            scheduler_queue_length=None,
    ):
        """

        :param seeds:
        :param project:
        :param task_name:
        :param redis_info:
        :param storer_info:
        :param scheduler_info: dict(DB="", table="", size="", config="")
        :param spider_num:
        :param max_retries:
        :param storer_queue_length:
        :param scheduler_queue_length:
        """
        self.seeds = struct_start_seeds(seeds)
        self.project = project or "test"
        self.task_name = task_name or "spider"

        self.oss_config = oss_config

        self.redis_info = parse_info(redis_info)
        self.storer_info = parse_info(storer_info)
        self.scheduler_info = parse_info(scheduler_info)

        self.spider_num = spider_num or 1
        self.max_retries = max_retries or 5
        self.storer_queue_length = storer_queue_length or 100
        self.scheduler_queue_length = scheduler_queue_length or 100




import time
import threading
from threading import Thread
from importlib import import_module

from cobweb import log, Queue, DBItem, RedisDB, OssDB, StorerInterface
from cobweb.utils import struct_queue_name, restore_table_name
from .models import Scheduler, Spider, Storer
# from collections import namedtuple

# def start_seeds(seeds):
#     if not seeds:
#         return None
#     if any(isinstance(seeds, t) for t in (list, tuple)):
#         return [Seed(seed) for seed in seeds]
#     elif any(isinstance(seeds, t) for t in (str, dict)):
#         return Seed(seeds)


# def parse_storer_info(storer_info):
#     storer_data = {}
#     storer_info_list = []
#     if storer_info.__class__.__name__ == 'StorerInfo':
#         storer_info_list.append(storer_info)
#     elif any(isinstance(storer_info, t) for t in (list, tuple)):
#         storer_info_list = storer_info
#     for info in storer_info_list:
#         db_name = info.DB.__name__
#         storer_data.setdefault(db_name, {"StorerDB": info.DB, "db_args_list": []})
#         storer_data[db_name]["db_args_list"].append(info[1:])
#     return storer_data

def get_scheduler_db(db):
    if isinstance(db, str):
        if "." in db:
            model_path = db.split(".")
            model = import_module(db)
            obj = getattr(model, db)
        else:
            model = import_module(f"cobweb.db.scheduler.{db.lower()}")
            obj = getattr(model, db.capitalize())
        return obj
        # if db.lower() in dir(StorerDB):
        #     return getattr(StorerDB, db)
        # else:
        #     pass
    elif issubclass(db, StorerInterface):
        return db
    raise TypeError()


def get_storer_db(db):
    if isinstance(db, str):
        if "." in db:
            model_path = db.split(".")
            model = import_module(db)
            obj = getattr(model, db)
        else:
            model = import_module(f"cobweb.db.storer.{db.lower()}")
            obj = getattr(model, db.capitalize())
        return obj, db.lower()
        # if db.lower() in dir(StorerDB):
        #     return getattr(StorerDB, db)
        # else:
        #     pass
    elif issubclass(db, StorerInterface):
        return db, db.__name__.lower()
    raise TypeError()


def check(stop, last, spider, scheduler, storer_list, ready_seed_length, spider_queue_length):
    time.sleep(30)
    spider_info = """
----------------------- check -----------------------
            running_spider_thread_num: {0}
            redis_ready_seed_length:   {1}
            redis_spider_seed_length:  {2}
            memory_seed_queue_length:  {3}
            storer_upload_queue_length_info: 
                {4}
-----------------------  end  -----------------------"""
    while True:

        running_spider_thread_num = spider.spider_in_progress.length
        redis_ready_seed_length = ready_seed_length()
        redis_spider_seed_length = spider_queue_length()
        memory_seed_queue_length = scheduler.queue.length
        storer_upload_queue_list = []
        storer_queue_empty = True
        for storer in storer_list:
            if storer.queue.length:
                storer_queue_empty = False
            storer_upload_queue_list.append(
                f"{storer.__class__.__name__}: {storer.queue.length}"
            )
        storer_upload_queue_length_info = "\n            ".join(storer_upload_queue_list)

        if (
                scheduler.stop and
                not redis_ready_seed_length and
                not memory_seed_queue_length and
                not running_spider_thread_num
        ):
            log.info("spider is done?")
            last.set()
            if storer_queue_empty and not redis_spider_seed_length:
                log.info("spider done!")
                break
        last.clear()

        log.info(spider_info.format(
            running_spider_thread_num,
            redis_ready_seed_length,
            redis_spider_seed_length,
            memory_seed_queue_length,
            storer_upload_queue_length_info
        ))

        time.sleep(3)
    stop.set()


def launcher(task):
    """
    任务启动装饰器
    :param task: 任务配置信息
    """
    def decorator(func):
        """
        Item:
            Textfile()
            Loghub()
            Console()
        e.g.
        task.fields = "a,b"
        func(item, seed)
            a = "a"
            b = "b"
            data = {"a": "a", "b": "b"}
            yield item.Loghub(**data)
            yield item.Loghub(a=a, b=b)
        """
        storer_list = []

        # 程序结束事件
        last = threading.Event()
        # 停止采集事件
        stop = threading.Event()

        # 初始化redis信息
        redis_db = RedisDB(task.project, task.task_name, task.redis_info)

        log.info("初始化cobweb!")

        seed_queue = Queue()

        if task.scheduler_info is None:
            task.scheduler_info = dict()

        # 调度器动态继承
        sql = task.scheduler_info.get("sql")
        table = task.scheduler_info.get("table")
        size = task.scheduler_info.get("size")
        scheduler_config = task.scheduler_info.get("config")
        scheduler_db = task.scheduler_info.get("db", "default")
        DB = get_scheduler_db(scheduler_db)
        # SchedulerDB, table, sql, length, size, config = task.scheduler_info
        SchedulerTmp = type(DB.__name__, (Scheduler, DB), {})

        # 初始化调度器
        scheduler = SchedulerTmp(
            table=table, sql=sql, size=size, queue=seed_queue,
            length=task.scheduler_queue_length, config=scheduler_config
        )

        # 初始化采集器
        spider = Spider(seed_queue, task.max_retries)

        # 解析存储器信息
        storer_info_list = task.storer_info
        if not isinstance(storer_info_list, list):
            storer_info_list = [storer_info_list]

        # new item
        item = type("Item", (object,), {"redis_client": redis_db})()

        if task.oss_config:
            item.oss = OssDB(**task.oss_config)

        for storer_info in storer_info_list:
            storer_db = storer_info["db"]
            fields = storer_info["fields"]
            storer_table = storer_info.get("table", "console")
            storer_config = storer_info.get("config")

            StorerDB, db_name = get_storer_db(storer_db)
            StorerTmp = type(StorerDB.__name__, (Storer, StorerDB), {})

            if not getattr(item, db_name, None):
                instance = type(db_name, (DBItem,), {})
                setattr(item, db_name, instance)

            storer_item_instance = getattr(item, db_name)
            storer_item_instance.init_item(storer_table, fields)

            storer_queue = struct_queue_name(db_name, storer_table)
            queue = getattr(storer_item_instance, storer_queue)
            # 初始话存储器
            table_name = restore_table_name(table_name=storer_table)
            storer = StorerTmp(
                table=table_name, fields=fields,
                length=task.storer_queue_length,
                queue=queue, config=storer_config
            )
            storer_list.append(storer)

        # for db_name in storer_data.keys():
        #     # 存储器动态继承
        #     StorerDB = storer_data[db_name]["StorerDB"]
        #     StorerTmp = type(StorerDB.__name__, (Storer, StorerDB), {})
        #     db_args_list = storer_data[db_name]["db_args_list"]
        #     for storer_db_args in db_args_list:
        #         table, fields, length, config = storer_db_args
        #         if not getattr(item, db_name, None):
        #             instance = type(db_name, (DBItem,), {})
        #             setattr(item, db_name, instance)
        #         # 创建存储xxx, 创建存储队列
        #         storer_item_instance = getattr(item, db_name)
        #         storer_item_instance.init_item(table, fields)
        #         #
        #         storer_queue = struct_queue_name(db_name, table)
        #         queue = getattr(storer_item_instance, storer_queue)
        #         # 初始话存储器
        #         table_name = restore_table_name(table_name=table)
        #         storer = StorerTmp(table_name, fields, length, queue, config)
        #         storer_list.append(storer)

        Thread(target=redis_db.check_spider_queue, args=(stop, len(storer_list))).start()
        Thread(target=redis_db.set_heartbeat, args=(stop,)).start()

        # 推送初始种子
        # seeds = start_seeds(task.start_seed)
        redis_db.add_seed(task.seeds)
        # 启动调度器, 调度至redis队列
        Thread(
            # name="xxxx_schedule_seeds",
            target=scheduler.schedule_seed,
            args=(
                redis_db.ready_seed_length,
                redis_db.get_scheduler_lock,
                redis_db.add_seed
            )
        ).start()

        # 启动调度器, 调度任务队列
        Thread(
            # name="xxxx_schedule_task",
            target=scheduler.schedule_task,
            args=(
                stop, redis_db.get_seed,
                redis_db.ready_seed_length
            )
        ).start()

        # 启动采集器
        for index in range(task.spider_num):
            Thread(
                # name=f"xxxx_spider_task:{index}",
                target=spider.spider_task,
                args=(
                    stop, func, item,
                    redis_db.del_seed
                )
            ).start()

        # 启动存储器
        for storer in storer_list:
            Thread(
                # name=f"xxxx_store_task:{storer.table}",
                target=storer.store_task,
                args=(
                    stop, last,
                    redis_db.reset_seed,
                    redis_db.set_storer
                )
            ).start()

        Thread(
            # name="check_spider",
            target=check,
            args=(
                stop, last, spider,
                scheduler, storer_list,
                redis_db.ready_seed_length,
                redis_db.spider_queue_length,
            )
        ).start()

    return decorator


# model = get_storer_db("console")
# print()


import json
from aliyun.log import LogClient, LogItem, PutLogsRequest
from cobweb import log, StorerInterface


class Loghub(StorerInterface):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.client = None

    def init_loghub_clint(self):
        try:
            self.client = LogClient(
                self.config['endpoint'],
                self.config['access_key_id'],
                self.config['access_key']
            )
        except Exception as e:
            self.client = None
            return False

    def store(self, data_list):
        try:
            if not self.client:
                self.init_loghub_clint()

            log_items = list()
            for item in data_list:
                temp = item._asdict()
                for key, value in temp.items():
                    if isinstance(value, str):
                        temp[key] = value
                    else:
                        temp[key] = json.dumps(value, ensure_ascii=False)
                log_item = LogItem()
                contents = sorted(temp.items())  # dict to tuple
                log_item.set_contents(contents)
                log_items.append(log_item)
            request = PutLogsRequest(
                project=self.config["project"],
                logstore=self.table,
                topic=self.config["topic"],
                source=self.config.get("source"),
                logitems=log_items,
                compress=True
            )
            self.client.put_logs(request=request)
            log.info(f"save data, data length: {len(data_list)}")
            return True
        except Exception as e:
            log.exception(e)
            return False




import time
import threading

from single.nest import Seed, DBItem
from single.nest import struct_queue_name, restore_table_name
from single.nest import Distributor, Scheduler, Spider, Storer


def init_task_seed(seeds):
    if not seeds:
        return None
    if isinstance(seeds, list) or isinstance(seeds, tuple):
        for seed in seeds:
            yield Seed(seed)
    elif isinstance(seeds, str) or isinstance(seeds, dict):
        yield Seed(seeds)


def parse_storer_info(storer_info):
    storer_data = {}
    storer_info_list = []
    if storer_info.__class__.__name__ == 'StorerInfo':
        storer_info_list.append(storer_info)
    elif isinstance(storer_info, tuple) or isinstance(storer_info, list):
        storer_info_list = storer_info
    for info in storer_info_list:
        db_name = info.DB.__name__
        storer_data.setdefault(db_name, {"StorerDB": info.DB, "db_args_list": []})
        storer_data[db_name]["db_args_list"].append(info[1:])
    return storer_data


def check(stop_event, last_event, distributor, scheduler, spider, storer_list):
    while True:
        time.sleep(3)
        if (
                scheduler.stop and
                not distributor.seed_queue.length and
                not spider.spider_in_progress.length
        ):
            last_event.set()
            time.sleep(10)
            storer_queue_empty = True
            for storer in storer_list:
                if storer.queue.length:
                    storer_queue_empty = False
                    break
            if storer_queue_empty:
                break
        last_event.clear()
    stop_event.set()


def cobweb(task):
    """
    任务启动装饰器
    :param task: 任务配置信息
    """
    def decorator(func):
        """
        func(Item, seed)
            Item:
                Item.Textfile()
                Item.Console()
        """
        # project task_name start_seed spider_num queue_length scheduler_info storer_info

        storer_list = []

        # 程序结束事件
        last_event = threading.Event()
        # 暂停采集事件
        stop_event = threading.Event()

        # 创建分发器
        distributor = Distributor()

        # 调度器动态继承
        SchedulerDB, table, sql, length, size = task.SchedulerInfo
        SchedulerTmp = type('Scheduler', (Scheduler, SchedulerDB), {})

        # 初始化调度器
        scheduler = SchedulerTmp(table=table, sql=sql, length=length, size=size, queue=distributor.seed_queue)

        # 初始化采集器
        spider = Spider(queue=distributor.seed_queue)

        # 解析存储器信息
        storer_data = parse_storer_info(task.storer_info)

        # sds
        item = type("item", (object,), {})
        for db_name in storer_data.keys():
            # 存储器动态继承
            StorerDB = storer_data[db_name]["StorerDB"]
            StorerTmp = type('Storer', (Storer, StorerDB), {})
            db_args_list = storer_data[db_name]["db_args_list"]
            for storer_db_args in db_args_list:
                table, fields, length = storer_db_args
                if not getattr(item, db_name, None):
                    instance = type(db_name, (DBItem,), {})
                    setattr(item, db_name, instance)
                # 创建存储xxx
                getattr(item, db_name).init_item(table, fields)
                # 创建存储队列
                storer_queue = struct_queue_name(db_name, table)
                distributor.create_queue(queue_name=storer_queue)
                queue = distributor.get_queue(queue_name=storer_queue)
                # 初始话存储器
                table_name = restore_table_name(table_name=table)
                storer = StorerTmp(table=table_name, fields=fields, length=length, queue=queue)
                storer_list.append(storer)

        # 推送初始种子
        distributor.distribute(init_task_seed, seeds=task.start_seed)

        # 启动调度器
        threading.Thread(
            target=scheduler.schedule_task,
            args=(distributor.distribute,),
            name="single_scheduler_task"
        ).start()

        # 启动采集器
        for index in range(task.spider_num):
            threading.Thread(
                target=spider.spider_task,
                args=(stop_event, distributor.distribute, func, item),
                name=f"single_spider_task:{index}"
            ).start()

        # 启动存储器
        for storer in storer_list:
            threading.Thread(
                target=storer.store_task,
                args=(stop_event, last_event, distributor.distribute),
                name=f"single_store_task:{storer.table}",
            ).start()

        threading.Thread(
            target=check, name="check",
            args=(
                stop_event, last_event, distributor,
                scheduler, spider, storer_list
            )
        ).start()

        # return starter(task, func)
    return decorator







# from typing import Iterable
import json
import time
import hashlib
from .log import log
from .utils import struct_queue_name
from collections import deque, namedtuple


class Queue:

    def __init__(self):
        self._queue = deque()

    @property
    def length(self) -> int:
        return len(self._queue)
    #
    # @property
    # def queue_names(self):
    #     return tuple(self.__dict__.keys())
    #
    # @property
    # def used_memory(self):
    #     return asizeof.asizeof(self)

    # def create_queue(self, queue_name: str):
    #     self.__setattr__(queue_name, deque())

    # def push_seed(self, seed):
    #     self.push("_seed_queue", seed)

    # def pop_seed(self):
    #     return self.pop("_seed_queue")

    def push(self, data, left: bool = False, direct_insertion: bool = False):
        try:
            if not data:
                return None
            if direct_insertion or isinstance(data, Seed):
                self._queue.appendleft(data) if left else self._queue.append(data)
            elif any(isinstance(data, t) for t in (list, tuple)):
                self._queue.extendleft(data) if left else self._queue.extend(data)
        except AttributeError as e:
            log.exception(e)

    def pop(self, left: bool = True):
        try:
            return self._queue.popleft() if left else self._queue.pop()
        except IndexError:
            return None
        except AttributeError as e:
            log.exception(e)
            return None


class Seed:

    def __init__(
            self,
            seed_info=None,
            priority=300,
            version=0,
            retry=0,
            **kwargs
    ):
        if seed_info:
            if any(isinstance(seed_info, t) for t in (str, bytes)):
                try:
                    item = json.loads(seed_info)
                    for k, v in item.items():
                        self.__setattr__(k, v)
                except json.JSONDecodeError:
                    self.__setattr__("url", seed_info)
            elif isinstance(seed_info, dict):
                for k, v in seed_info.items():
                    self.__setattr__(k, v)
            else:
                raise TypeError(Exception(
                    f"seed type error, "
                    f"must be str or dict! "
                    f"seed_info: {seed_info}"
                ))
        for k, v in kwargs.items():
            self.__setattr__(k, v)
        if not getattr(self, "_priority"):
            self._priority = min(max(1, int(priority)), 999)
        if not getattr(self, "_version"):
            self._version = int(version) or int(time.time())
        if not getattr(self, "_retry"):
            self._retry = retry
        if not getattr(self, "sid"):
            self.init_id()

    def init_id(self):
        item_string = self.format_seed
        seed_id = hashlib.md5(item_string.encode()).hexdigest()
        self.__setattr__("sid", seed_id)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, item):
        return getattr(self, item)

    def __getattr__(self, name):
        return None

    def __str__(self):
        return json.dumps(self.__dict__, ensure_ascii=False)

    def __repr__(self):
        chars = [f"{k}={v}" for k, v in self.__dict__.items()]
        return f'{self.__class__.__name__}({", ".join(chars)})'

    @property
    def format_seed(self):
        seed = self.__dict__.copy()
        del seed["_priority"]
        del seed["_version"]
        del seed["_retry"]
        return json.dumps(seed, ensure_ascii=False)


class DBItem:

    def __init__(self, **kwargs):
        self.__setattr__("_index", 0, True)
        for table in self.__class__.__table__:
            if set(kwargs.keys()) == set(table._fields):
                break
            self._index += 1

        if self._index > len(self.__class__.__table__):
            raise Exception()

        table = self.__class__.__table__[self._index]
        self.__setattr__("struct_data", table(**kwargs), True)
        self.__setattr__("db_name", self.__class__.__name__, True)
        self.__setattr__("table_name", self.struct_data.__class__.__name__, True)

    @classmethod
    def init_item(cls, table_name, fields):
        queue_name = struct_queue_name(cls.__name__, table_name)
        if getattr(cls, queue_name, None) is None:
            setattr(cls, queue_name, Queue())

        if getattr(cls, "__table__", None) is None:
            cls.__table__ = []

        table = namedtuple(table_name, fields)

        if table in getattr(cls, "__table__"):
            raise Exception()
        getattr(cls, "__table__").append(table)

    def queue(self):
        queue_name = struct_queue_name(self.db_name, self.table_name)
        return getattr(self.__class__, queue_name)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, item):
        return self.struct_data[item]

    def __getattr__(self, name):
        return None

    def __setattr__(self, key, value, init=None):
        if init:
            super().__setattr__(key, value)
        elif not getattr(self, "struct_data"):
            raise Exception(f"no struct_data")
        else:
            self.__setattr__(
                "struct_data",
                self.struct_data._replace(**{key: value}),
                init=True
            )

    def __str__(self):
        return json.dumps(self.struct_data._asdict(), ensure_ascii=False)

    def __repr__(self):
        return f'{self.__class__.__name__}:{self.struct_data}'




from cobweb import log, StorerInterface


class Redis(StorerInterface):

    def store(self, data_list):
        try:
            data_str = "\n".join(str(data) for data in data_list)
            with open(self.table, "a") as fp:
                fp.write(data_str)
            log.info(f"save data, data length: {len(data_list)}")
            return True
        except Exception as e:
            return False




from functools import wraps
from cobweb import log


def check_redis_status(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            log.exception(e)
            result = False
        return result

    return wrapper




import json
import sys
from typing import Iterable

import requests


# from cobweb import Seed


def struct_table_name(table_name):
    return table_name.replace(".", "__p__").replace(":", "__c__")


def restore_table_name(table_name):
    return table_name.replace("__p__", ".").replace("__c__", ":")


def struct_queue_name(db_name, table_name):
    return sys.intern(f"__{db_name}_{table_name}_queue__")


# class StorerDB:
#
#     @staticmethod
#     def console(self):
#         from db.storer.console import Console
#         table = struct_table_name(table)
#         return StorerInfo(DB=Console, table=table, length=length, config=None)
#
#     @staticmethod
#     def textfile(table, length=200):
#         from db.storer.textfile import Textfile
#         table = struct_table_name(table)
#         return StorerInfo(DB=Textfile, table=table, length=length, config=None)
#
#     @staticmethod
#     def loghub(table, length=200, config=None):
#         from db.storer.loghub import Loghub
#         table = struct_table_name(table)
#         return StorerInfo(DB=Loghub, table=table, length=length, config=config)


def parse_info(info):
    if not info:
        return info

    if isinstance(info, dict):
        return info

    if isinstance(info, str):
        return json.loads(info)

    if isinstance(info, Iterable):
        result = list()
        for ii in info:
            if isinstance(ii, str):
                result.append(json.loads(ii))
            elif isinstance(ii, dict):
                result.append(ii)
            else:
                raise TypeError("must be in [str, dict]")

        return result


def struct_start_seeds(seeds):
    from .bbb import Seed
    if not seeds:
        return None
    if any(isinstance(seeds, t) for t in (list, tuple)):
        return [Seed(seed) for seed in seeds]
    elif any(isinstance(seeds, t) for t in (str, dict)):
        return Seed(seeds)


# def get_storer_db(db):
#
#     if isinstance(db, str):
#         model = import_module(f" db.storer.{db.lower()}")
#
#         # if db.lower() in dir(StorerDB):
#         #     return getattr(StorerDB, db)
#         # else:
#         #     pass






from abc import ABC, abstractmethod
from .utils import parse_info


class SchedulerInterface(ABC):

    def __init__(self, table, sql, length, size, queue, config=None):
        self.sql = sql
        self.table = table
        self.length = length
        self.size = size
        self.queue = queue
        self.config = parse_info(config)
        self.stop = False

    @abstractmethod
    def schedule(self, *args, **kwargs):
        pass


class StorerInterface(ABC):

    def __init__(self, table, fields, length, queue, config=None):
        self.table = table
        self.fields = fields
        self.length = length
        self.queue = queue
        self.config = parse_info(config)
        # self.redis_db = redis_db

    @abstractmethod
    def store(self, *args, **kwargs):
        pass

